"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SolanaPlatform = void 0;
const web3_js_1 = require("@solana/web3.js");
const connect_sdk_1 = require("@wormhole-foundation/connect-sdk");
const contracts_1 = require("./contracts");
const chain_1 = require("./chain");
const tokenBridge_1 = require("./protocols/tokenBridge");
const platformUtils_1 = require("./platformUtils");
const SOLANA_SEQ_LOG = 'Program log: Sequence: ';
// forces SolanaPlatform to implement Platform
var _ = SolanaPlatform;
/**
 * @category Solana
 */
var SolanaPlatform;
(function (SolanaPlatform) {
    SolanaPlatform.platform = 'Solana';
    SolanaPlatform.network = connect_sdk_1.DEFAULT_NETWORK;
    SolanaPlatform.conf = (0, connect_sdk_1.networkPlatformConfigs)(SolanaPlatform.network, SolanaPlatform.platform);
    let contracts = new contracts_1.SolanaContracts(SolanaPlatform.conf);
    SolanaPlatform.nativeTokenId = platformUtils_1.SolanaUtils.nativeTokenId, SolanaPlatform.isNativeTokenId = platformUtils_1.SolanaUtils.isNativeTokenId, SolanaPlatform.isSupportedChain = platformUtils_1.SolanaUtils.isSupportedChain, SolanaPlatform.getDecimals = platformUtils_1.SolanaUtils.getDecimals, SolanaPlatform.getBalance = platformUtils_1.SolanaUtils.getBalance, SolanaPlatform.sendWait = platformUtils_1.SolanaUtils.sendWait, SolanaPlatform.getCurrentBlock = platformUtils_1.SolanaUtils.getCurrentBlock, SolanaPlatform.chainFromRpc = platformUtils_1.SolanaUtils.chainFromRpc;
    function setConfig(network, _conf) {
        SolanaPlatform.conf = _conf ? _conf : (0, connect_sdk_1.networkPlatformConfigs)(network, SolanaPlatform.platform);
        contracts = new contracts_1.SolanaContracts(SolanaPlatform.conf);
        return SolanaPlatform;
    }
    SolanaPlatform.setConfig = setConfig;
    function getRpc(chain, commitment = 'confirmed') {
        const rpcAddress = SolanaPlatform.conf[chain].rpc;
        return new web3_js_1.Connection(rpcAddress, commitment);
    }
    SolanaPlatform.getRpc = getRpc;
    function getChain(chain) {
        return new chain_1.SolanaChain(chain);
    }
    SolanaPlatform.getChain = getChain;
    async function getTokenBridge(rpc) {
        return tokenBridge_1.SolanaTokenBridge.fromProvider(rpc, contracts);
    }
    SolanaPlatform.getTokenBridge = getTokenBridge;
    async function parseTransaction(chain, rpc, tx) {
        const _contracts = contracts.getContracts(chain);
        if (!_contracts.coreBridge)
            throw new Error('contracts not found');
        const response = await rpc.getTransaction(tx);
        if (!response || !response.meta?.innerInstructions[0].instructions)
            throw new Error('transaction not found');
        const instructions = response.meta?.innerInstructions[0].instructions;
        const accounts = response.transaction.message.accountKeys;
        // find the instruction where the programId equals the Wormhole ProgramId and the emitter equals the Token Bridge
        const bridgeInstructions = instructions.filter((i) => {
            const programId = accounts[i.programIdIndex].toString();
            const wormholeCore = _contracts.coreBridge;
            return programId === wormholeCore;
        });
        if (bridgeInstructions.length === 0)
            throw new Error('no bridge messages found');
        // TODO: unsure about the single bridge instruction and the [2] index, will this always be the case?
        const [logmsg] = bridgeInstructions;
        const emitterAcct = accounts[logmsg.accounts[2]];
        const emitter = (0, connect_sdk_1.toNative)(chain, emitterAcct.toString());
        const sequence = response.meta?.logMessages
            ?.filter((msg) => msg.startsWith(SOLANA_SEQ_LOG))?.[0]
            ?.replace(SOLANA_SEQ_LOG, '');
        if (!sequence) {
            throw new Error('sequence not found');
        }
        return [
            {
                chain,
                emitter: emitter.toUniversalAddress(),
                sequence: BigInt(sequence),
            },
        ];
    }
    SolanaPlatform.parseTransaction = parseTransaction;
})(SolanaPlatform || (exports.SolanaPlatform = SolanaPlatform = {}));
//# sourceMappingURL=platform.js.map