"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SolanaTokenBridge = void 0;
const connect_sdk_1 = require("@wormhole-foundation/connect-sdk");
const web3_js_1 = require("@solana/web3.js");
const spl_token_1 = require("@solana/spl-token");
const wormhole_1 = require("../utils/wormhole");
const tokenBridge_1 = require("../utils/tokenBridge");
const unsignedTransaction_1 = require("../unsignedTransaction");
const platform_1 = require("../platform");
class SolanaTokenBridge {
    network;
    chain;
    connection;
    contracts;
    chainId;
    tokenBridge;
    coreBridge;
    constructor(network, chain, connection, contracts) {
        this.network = network;
        this.chain = chain;
        this.connection = connection;
        this.contracts = contracts;
        this.chainId = (0, connect_sdk_1.toChainId)(chain);
        this.tokenBridge = this.contracts.getTokenBridge(chain, connection);
        this.coreBridge = this.contracts.getCore(chain, connection);
    }
    static async fromProvider(connection, contracts) {
        const [network, chain] = await platform_1.SolanaPlatform.chainFromRpc(connection);
        return new SolanaTokenBridge(network, chain, connection, contracts);
    }
    async isWrappedAsset(token) {
        return (0, tokenBridge_1.getWrappedMeta)(this.connection, this.tokenBridge.programId, token.toUint8Array())
            .catch((_) => null)
            .then((meta) => meta != null);
    }
    async getOriginalAsset(token) {
        if (!(await this.isWrappedAsset(token)))
            throw (0, connect_sdk_1.ErrNotWrapped)(token.toString());
        const mint = new web3_js_1.PublicKey(token.toUint8Array());
        try {
            const meta = await (0, tokenBridge_1.getWrappedMeta)(this.connection, this.tokenBridge.programId, token.toUint8Array());
            if (meta === null)
                return {
                    chain: this.chain,
                    address: (0, connect_sdk_1.toNative)(this.chain, mint.toBytes()),
                };
            return {
                chain: (0, connect_sdk_1.toChainName)(meta.chain),
                address: new connect_sdk_1.UniversalAddress(meta.tokenAddress),
            };
        }
        catch (_) {
            // TODO: https://github.com/wormhole-foundation/wormhole/blob/main/sdk/js/src/token_bridge/getOriginalAsset.ts#L200
            // the current one returns 0s for address
            throw (0, connect_sdk_1.ErrNotWrapped)(token.toString());
        }
    }
    async hasWrappedAsset(token) {
        try {
            await this.getWrappedAsset(token);
            return true;
        }
        catch (_) { }
        return false;
    }
    async getWrappedAsset(token) {
        const mint = (0, tokenBridge_1.deriveWrappedMintKey)(this.tokenBridge.programId, (0, connect_sdk_1.toChainId)(token.chain), token.address.toUniversalAddress().toUint8Array());
        // If we don't throw an error getting wrapped meta, we're good to return
        // the derived mint address back to the caller.
        try {
            await (0, tokenBridge_1.getWrappedMeta)(this.connection, this.tokenBridge.programId, mint);
            return (0, connect_sdk_1.toNative)(this.chain, mint.toBase58());
        }
        catch (_) { }
        throw (0, connect_sdk_1.ErrNotWrapped)(token.address.toUniversalAddress().toString());
    }
    async isTransferCompleted(vaa) {
        throw new Error('not implemented');
    }
    async *createAttestation(token, sender) {
        const senderAddress = new web3_js_1.PublicKey(sender.toUint8Array());
        // TODO:
        const nonce = 0; // createNonce().readUInt32LE(0);
        const transferIx = await (0, wormhole_1.createBridgeFeeTransferInstruction)(this.connection, this.coreBridge.programId, senderAddress);
        const messageKey = web3_js_1.Keypair.generate();
        const attestIx = (0, tokenBridge_1.createAttestTokenInstruction)(this.connection, this.tokenBridge.programId, this.coreBridge.programId, senderAddress, token.toUint8Array(), messageKey.publicKey, nonce);
        const transaction = new web3_js_1.Transaction().add(transferIx, attestIx);
        const { blockhash } = await this.connection.getLatestBlockhash();
        transaction.recentBlockhash = blockhash;
        transaction.feePayer = senderAddress;
        transaction.partialSign(messageKey);
        yield this.createUnsignedTx(transaction, 'Solana.AttestToken');
    }
    async *submitAttestation(vaa, sender) {
        const senderAddress = new web3_js_1.PublicKey(sender.toUint8Array());
        const transaction = new web3_js_1.Transaction().add((0, tokenBridge_1.createCreateWrappedInstruction)(this.connection, this.tokenBridge.programId, this.coreBridge.programId, senderAddress, vaa));
        const { blockhash } = await this.connection.getLatestBlockhash();
        transaction.recentBlockhash = blockhash;
        transaction.feePayer = senderAddress;
        yield this.createUnsignedTx(transaction, 'Solana.CreateWrapped');
    }
    async transferSol(sender, recipient, amount, payload) {
        //  https://github.com/wormhole-foundation/wormhole-connect/blob/development/sdk/src/contexts/solana/context.ts#L245
        const senderAddress = new web3_js_1.PublicKey(sender.toUint8Array());
        // TODO: why?
        const payerPublicKey = senderAddress;
        const recipientAddress = recipient.address
            .toUniversalAddress()
            .toUint8Array();
        const recipientChainId = (0, connect_sdk_1.toChainId)(recipient.chain);
        const nonce = 0;
        const relayerFee = 0n;
        const message = web3_js_1.Keypair.generate();
        const ancillaryKeypair = web3_js_1.Keypair.generate();
        const rentBalance = await (0, spl_token_1.getMinimumBalanceForRentExemptAccount)(this.connection);
        //This will create a temporary account where the wSOL will be created.
        const createAncillaryAccountIx = web3_js_1.SystemProgram.createAccount({
            fromPubkey: payerPublicKey,
            newAccountPubkey: ancillaryKeypair.publicKey,
            lamports: rentBalance,
            space: spl_token_1.ACCOUNT_SIZE,
            programId: spl_token_1.TOKEN_PROGRAM_ID,
        });
        //Send in the amount of SOL which we want converted to wSOL
        const initialBalanceTransferIx = web3_js_1.SystemProgram.transfer({
            fromPubkey: payerPublicKey,
            lamports: amount,
            toPubkey: ancillaryKeypair.publicKey,
        });
        //Initialize the account as a WSOL account, with the original payerAddress as owner
        const initAccountIx = (0, spl_token_1.createInitializeAccountInstruction)(ancillaryKeypair.publicKey, spl_token_1.NATIVE_MINT, payerPublicKey);
        //Normal approve & transfer instructions, except that the wSOL is sent from the ancillary account.
        const approvalIx = (0, tokenBridge_1.createApproveAuthoritySignerInstruction)(this.tokenBridge.programId, ancillaryKeypair.publicKey, payerPublicKey, amount);
        const tokenBridgeTransferIx = payload
            ? (0, tokenBridge_1.createTransferNativeWithPayloadInstruction)(this.connection, this.tokenBridge.programId, this.coreBridge.programId, senderAddress, message.publicKey, ancillaryKeypair.publicKey, spl_token_1.NATIVE_MINT, nonce, amount, recipientAddress, recipientChainId, payload)
            : (0, tokenBridge_1.createTransferNativeInstruction)(this.connection, this.tokenBridge.programId, this.coreBridge.programId, senderAddress, message.publicKey, ancillaryKeypair.publicKey, spl_token_1.NATIVE_MINT, nonce, amount, relayerFee, recipientAddress, recipientChainId);
        //Close the ancillary account for cleanup. Payer address receives any remaining funds
        const closeAccountIx = (0, spl_token_1.createCloseAccountInstruction)(ancillaryKeypair.publicKey, //account to close
        payerPublicKey, //Remaining funds destination
        payerPublicKey);
        const { blockhash } = await this.connection.getLatestBlockhash();
        const transaction = new web3_js_1.Transaction();
        transaction.recentBlockhash = blockhash;
        transaction.feePayer = payerPublicKey;
        transaction.add(createAncillaryAccountIx, initialBalanceTransferIx, initAccountIx, approvalIx, tokenBridgeTransferIx, closeAccountIx);
        transaction.partialSign(message, ancillaryKeypair);
        return this.createUnsignedTx(transaction, 'Solana.TransferNative');
    }
    async *transfer(sender, recipient, token, amount, payload) {
        // TODO: payer vs sender?? can caller add diff payer later?
        if (token === 'native') {
            yield await this.transferSol(sender, recipient, amount, payload);
            return;
        }
        const tokenAddress = new web3_js_1.PublicKey(token.toUint8Array());
        const senderAddress = new web3_js_1.PublicKey(sender.toUint8Array());
        const senderTokenAddress = await (0, spl_token_1.getAssociatedTokenAddress)(tokenAddress, senderAddress);
        const recipientAddress = recipient.address
            .toUniversalAddress()
            .toUint8Array();
        const recipientChainId = (0, connect_sdk_1.toChainId)(recipient.chain);
        const nonce = 0;
        const relayerFee = 0n;
        const isSolanaNative = !(await this.isWrappedAsset(token));
        const message = web3_js_1.Keypair.generate();
        let tokenBridgeTransferIx;
        if (isSolanaNative) {
            tokenBridgeTransferIx = payload
                ? (0, tokenBridge_1.createTransferNativeWithPayloadInstruction)(this.connection, this.tokenBridge.programId, this.coreBridge.programId, senderAddress, message.publicKey, senderTokenAddress, tokenAddress, nonce, amount, recipientAddress, recipientChainId, payload)
                : (0, tokenBridge_1.createTransferNativeInstruction)(this.connection, this.tokenBridge.programId, this.coreBridge.programId, senderAddress, message.publicKey, senderTokenAddress, tokenAddress, nonce, amount, relayerFee, recipientAddress, recipientChainId);
        }
        else {
            const originAsset = await this.getOriginalAsset(token);
            tokenBridgeTransferIx = payload
                ? (0, tokenBridge_1.createTransferWrappedWithPayloadInstruction)(this.connection, this.tokenBridge.programId, this.coreBridge.programId, senderAddress, message.publicKey, senderTokenAddress, senderAddress, (0, connect_sdk_1.toChainId)(originAsset.chain), originAsset.address.toUint8Array(), nonce, amount, recipientAddress, recipientChainId, payload)
                : (0, tokenBridge_1.createTransferWrappedInstruction)(this.connection, this.tokenBridge.programId, this.coreBridge.programId, senderAddress, message.publicKey, senderTokenAddress, senderAddress, (0, connect_sdk_1.toChainId)(originAsset.chain), originAsset.address.toUint8Array(), nonce, amount, relayerFee, recipientAddress, recipientChainId);
        }
        const approvalIx = (0, tokenBridge_1.createApproveAuthoritySignerInstruction)(this.tokenBridge.programId, token.toUint8Array(), senderAddress, amount);
        const transaction = new web3_js_1.Transaction().add(approvalIx, tokenBridgeTransferIx);
        const { blockhash } = await this.connection.getLatestBlockhash();
        transaction.recentBlockhash = blockhash;
        transaction.feePayer = senderAddress;
        transaction.partialSign(message);
        yield this.createUnsignedTx(transaction, 'Solana.TransferTokens');
    }
    async postVaa(sender, vaa) {
        const senderAddr = new web3_js_1.PublicKey(sender.toUint8Array());
        const signatureSet = web3_js_1.Keypair.generate();
        const verifySignaturesInstructions = await (0, wormhole_1.createVerifySignaturesInstructions)(this.connection, this.coreBridge.programId, senderAddr, vaa, signatureSet.publicKey);
        const unsignedTransactions = [];
        for (let i = 0; i < verifySignaturesInstructions.length; i += 2) {
            unsignedTransactions.push(new web3_js_1.Transaction().add(...verifySignaturesInstructions.slice(i, i + 2)));
        }
        unsignedTransactions.push(new web3_js_1.Transaction().add((0, wormhole_1.createPostVaaInstruction)(this.connection, this.coreBridge.programId, senderAddr, vaa, signatureSet.publicKey)));
        return {
            unsignedTransactions,
            signers: [signatureSet],
        };
    }
    async *redeem(sender, vaa, unwrapNative = true) {
        // TODO unwrapNative?
        // TODO: check if vaa.payload.token.address is native Sol
        const { blockhash } = await this.connection.getLatestBlockhash();
        const senderAddress = new web3_js_1.PublicKey(sender.toUint8Array());
        const ataAddress = new web3_js_1.PublicKey(vaa.payload.to.address.toUint8Array());
        const wrappedToken = await this.getWrappedAsset(vaa.payload.token);
        // If the ata doesn't exist yet, create it
        const acctInfo = await this.connection.getAccountInfo(ataAddress);
        if (acctInfo === null) {
            const ataCreationTx = new web3_js_1.Transaction().add((0, spl_token_1.createAssociatedTokenAccountInstruction)(senderAddress, ataAddress, senderAddress, new web3_js_1.PublicKey(wrappedToken.toUint8Array())));
            ataCreationTx.feePayer = senderAddress;
            ataCreationTx.recentBlockhash = blockhash;
            yield this.createUnsignedTx(ataCreationTx, 'Redeem.CreateATA');
        }
        // Get transactions to verify sigs and post the VAA
        const { unsignedTransactions: postVaaTxns, signers: postVaaSigners } = await this.postVaa(sender, vaa);
        // Take off the last tx in the set of postVAA txns
        // to send after verify sig txns
        const postVaaTx = postVaaTxns.pop();
        for (let i = 0; i < postVaaTxns.length; i++) {
            const verifySigTx = postVaaTxns[i];
            verifySigTx.recentBlockhash = blockhash;
            verifySigTx.feePayer = senderAddress;
            verifySigTx.partialSign(postVaaSigners[0]);
            yield this.createUnsignedTx(verifySigTx, 'Redeem.VerifySignature', 
            // all stackable except the last one
            // so we flush the buffer of sig verifies
            // and finalize prior to trying to Post the VAA
            i < postVaaTxns.length - 1);
        }
        postVaaTx.recentBlockhash = blockhash;
        postVaaTx.feePayer = senderAddress;
        yield this.createUnsignedTx(postVaaTx, 'Redeem.PostVAA');
        const createCompleteTransferInstruction = vaa.payload.token.chain == this.chain
            ? tokenBridge_1.createCompleteTransferNativeInstruction
            : tokenBridge_1.createCompleteTransferWrappedInstruction;
        const transaction = new web3_js_1.Transaction().add(createCompleteTransferInstruction(this.connection, this.tokenBridge.programId, this.coreBridge.programId, senderAddress, vaa));
        transaction.recentBlockhash = blockhash;
        transaction.feePayer = senderAddress;
        yield this.createUnsignedTx(transaction, 'Solana.RedeemTransfer');
    }
    async getWrappedNative() {
        return (0, connect_sdk_1.toNative)(this.chain, spl_token_1.NATIVE_MINT.toBase58());
    }
    createUnsignedTx(txReq, description, parallelizable = false) {
        return new unsignedTransaction_1.SolanaUnsignedTransaction(txReq, this.network, 'Solana', description, parallelizable);
    }
}
exports.SolanaTokenBridge = SolanaTokenBridge;
//# sourceMappingURL=tokenBridge.js.map