import { Connection } from '@solana/web3.js';
import { toNative, networkPlatformConfigs, DEFAULT_NETWORK, } from '@wormhole-foundation/connect-sdk';
import { SolanaContracts } from './contracts';
import { SolanaChain } from './chain';
import { SolanaTokenBridge } from './protocols/tokenBridge';
import { SolanaUtils } from './platformUtils';
const SOLANA_SEQ_LOG = 'Program log: Sequence: ';
// forces SolanaPlatform to implement Platform
var _ = SolanaPlatform;
/**
 * @category Solana
 */
export var SolanaPlatform;
(function (SolanaPlatform) {
    SolanaPlatform.platform = 'Solana';
    SolanaPlatform.network = DEFAULT_NETWORK;
    SolanaPlatform.conf = networkPlatformConfigs(SolanaPlatform.network, SolanaPlatform.platform);
    let contracts = new SolanaContracts(SolanaPlatform.conf);
    SolanaPlatform.nativeTokenId = SolanaUtils.nativeTokenId, SolanaPlatform.isNativeTokenId = SolanaUtils.isNativeTokenId, SolanaPlatform.isSupportedChain = SolanaUtils.isSupportedChain, SolanaPlatform.getDecimals = SolanaUtils.getDecimals, SolanaPlatform.getBalance = SolanaUtils.getBalance, SolanaPlatform.sendWait = SolanaUtils.sendWait, SolanaPlatform.getCurrentBlock = SolanaUtils.getCurrentBlock, SolanaPlatform.chainFromRpc = SolanaUtils.chainFromRpc;
    function setConfig(network, _conf) {
        SolanaPlatform.conf = _conf ? _conf : networkPlatformConfigs(network, SolanaPlatform.platform);
        contracts = new SolanaContracts(SolanaPlatform.conf);
        return SolanaPlatform;
    }
    SolanaPlatform.setConfig = setConfig;
    function getRpc(chain, commitment = 'confirmed') {
        const rpcAddress = SolanaPlatform.conf[chain].rpc;
        return new Connection(rpcAddress, commitment);
    }
    SolanaPlatform.getRpc = getRpc;
    function getChain(chain) {
        return new SolanaChain(chain);
    }
    SolanaPlatform.getChain = getChain;
    async function getTokenBridge(rpc) {
        return SolanaTokenBridge.fromProvider(rpc, contracts);
    }
    SolanaPlatform.getTokenBridge = getTokenBridge;
    async function parseTransaction(chain, rpc, tx) {
        const _contracts = contracts.getContracts(chain);
        if (!_contracts.coreBridge)
            throw new Error('contracts not found');
        const response = await rpc.getTransaction(tx);
        if (!response || !response.meta?.innerInstructions[0].instructions)
            throw new Error('transaction not found');
        const instructions = response.meta?.innerInstructions[0].instructions;
        const accounts = response.transaction.message.accountKeys;
        // find the instruction where the programId equals the Wormhole ProgramId and the emitter equals the Token Bridge
        const bridgeInstructions = instructions.filter((i) => {
            const programId = accounts[i.programIdIndex].toString();
            const wormholeCore = _contracts.coreBridge;
            return programId === wormholeCore;
        });
        if (bridgeInstructions.length === 0)
            throw new Error('no bridge messages found');
        // TODO: unsure about the single bridge instruction and the [2] index, will this always be the case?
        const [logmsg] = bridgeInstructions;
        const emitterAcct = accounts[logmsg.accounts[2]];
        const emitter = toNative(chain, emitterAcct.toString());
        const sequence = response.meta?.logMessages
            ?.filter((msg) => msg.startsWith(SOLANA_SEQ_LOG))?.[0]
            ?.replace(SOLANA_SEQ_LOG, '');
        if (!sequence) {
            throw new Error('sequence not found');
        }
        return [
            {
                chain,
                emitter: emitter.toUniversalAddress(),
                sequence: BigInt(sequence),
            },
        ];
    }
    SolanaPlatform.parseTransaction = parseTransaction;
})(SolanaPlatform || (SolanaPlatform = {}));
//# sourceMappingURL=platform.js.map